__version__ = '0.3.4'

from .client_template import (
    client_point_of_execution_f,
    client_method_template_f,
    client_class_def_template_f,
    client_imports_f
)

from .make_client import (
    main,
    convert_to_snake_case,
    create_client_endpoints,
    seems_like_a_url
)





if __name__ == '__main__':
    import argparse
    from swagccg.src.make_client import main as make_client

    parser = argparse.ArgumentParser()
    parser.add_argument('--config-path', '-c', action='store', default='config.json')
    args = parser.parse_args()
    make_client(config_path=args.config_path)



try:
    from .src import (
        client_point_of_execution_f,
        client_method_template_f,
        client_class_def_template_f,
        client_imports_f
    )
    from .src import (
        main,
        convert_to_snake_case,
        create_client_endpoints,
        seems_like_a_url
    )
except ImportError:
    from swagccg.src import (
        client_point_of_execution_f,
        client_method_template_f,
        client_class_def_template_f,
        client_imports_f
    )

    from swagccg.src import (
        main,
        convert_to_snake_case,
        create_client_endpoints,
        seems_like_a_url
    )
    print('Alternative import used')





import re
import json
import argparse
from urllib.parse import urlparse
from urllib.request import urlopen
try:
    from swagccg.src.client_template import client_imports_f
    from swagccg.src.client_template import client_class_def_template_f
    from swagccg.src.client_template import client_method_template_f
    from swagccg.src.client_template import client_point_of_execution_f
    from swagccg.src.client_template import client_encode_decoding_point_f
    from swagccg.src.client_template import dir_template_f
except ImportError:
    from .client_template import client_imports_f
    from .client_template import client_class_def_template_f
    from .client_template import client_method_template_f
    from .client_template import client_point_of_execution_f
    from .client_template import client_encoding_decoding_point_f
    from .client_template import dir_template_f
    print("Alternative import used.")

PARSED_HTTP_METHODS = ['GET', 'POST', 'DELETE', 'PATCH', 'PUT']


def convert_to_snake_case(name: str) -> str:
    """via https://stackoverflow.com/a/1176023 """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def create_client_endpoints(swagger_data, api_paths):
    methods_list = []
    method_names = []
    n = 0
    for api_path in api_paths:
        # compose the method name
        http_methods = list(swagger_data['paths'][api_path])

        for http_method in http_methods:
            operation = f'initial_{n}'
            n += 1
            doc_string = f' '
            if 'operationId' in swagger_data['paths'][api_path][http_method]:
                operation = swagger_data['paths'][api_path][http_method]['operationId']
            if 'summary' in swagger_data['paths'][api_path][http_method]:
                doc_string = swagger_data['paths'][api_path][http_method]['summary']

            # convert to snake case if it isn't already
            operation = convert_to_snake_case(operation)

            # check if the HTTP verb is already part of the operation name
            scan = [re.match(verb, operation, flags=re.IGNORECASE) is None for verb in PARSED_HTTP_METHODS]
            if len(scan) == sum(scan):
                method_name = f'{http_method}_{operation}_r'.lower()
            else:
                method_name = f'{operation}_r'.lower()
            method_name = method_name.replace('-', '_')
            method_names.append(method_name)
            # handle path parameters explicitly as opposed to parameters within the query string
            path_params = ''
            if 'parameters' in swagger_data['paths'][api_path][http_method]:
                parameters = swagger_data['paths'][api_path][http_method]['parameters']
                for p in parameters:
                    if p['in'] == 'path':
                        path_params += f', {p["name"]}'

            methods_list.append(
                client_method_template_f(
                    method_name=method_name,
                    http_verb=http_method,
                    api_path=api_path,
                    doc_string=doc_string,
                    path_params=path_params
                )
            )
    client_methods = ''
    for i in range(len(methods_list)):
        client_methods = client_methods + methods_list[i]
    return client_methods, method_names


def seems_like_a_url(url):
    """ is_url https://stackoverflow.com/a/52455972/10124294 """
    try:
        r = urlparse(url)
        return all([r.scheme, r.netloc, r.path])
    except ValueError:
        return False


def main(config_path=None):
    """ control flow """
    if config_path is None:
        config_path = 'config.json'
    if config_path is None:
        config_path = '/tmp/auto_client.py'
    try:
        with open(config_path, 'r') as f:
            config = json.load(f)
    except FileNotFoundError:
        msg = (f'Could not find a configuration file at: '
               f'\n{config_path}\n '
               f'Did you pass a path to the location of the configuration file?'
               f'The default assumption search only the working directory.'
               f'If yes, try to pass an absolute path.')
        raise FileNotFoundError(msg)

    try:
        if seems_like_a_url(config['swagger_path']):
            with urlopen(config['swagger_path']) as url:
                swagger_data = json.loads(url.read().decode())
        else:
            with open(config['swagger_path'], 'r') as f:
                swagger_data = json.load(f)
    except FileNotFoundError:
        msg = (f'Could not find the swagger specification file at:\n'
               f'{config["swagger_path"]}\n'
               f'A look inside {config_path} is worth a shot.')
        raise FileNotFoundError(msg)

    if 'basePath' in swagger_data:
        config['basePath'] = swagger_data['basePath']
    else:
        config['basePath'] = ''
        Warning(f"``basePath`` does not seem to be assigned within:\n"
                f"{config['swagger_path']}\n"
                f"Did not result in any data. This isn't necessarily a problem\n"
                f"Setting ``basePath`` as '' (empty string)")

    if 'paths' in swagger_data:
        api_paths = list(swagger_data['paths'])
    else:
        raise ValueError(f"The swagger definition below did not provide any ``paths``:\n"
                         f"{config['swagger_path']}\n"
                         f"An API definition without paths? Test? Exiting.")

    client_imports = client_imports_f()
    client_class_def = client_class_def_template_f(args=config)
    client_encode_decoding_point = client_encoding_decoding_point_f()
    client_point_of_execution = client_point_of_execution_f()

    client_methods, method_names = create_client_endpoints(
        api_paths=api_paths,
        swagger_data=swagger_data
    )
    client_dir_function_str = dir_template_f(method_names)
    client_class_def = client_class_def.replace('\n    # def __dir__(self):\n', client_dir_function_str, 1)
    all_in_one = (
            client_imports
            + client_class_def
            + client_point_of_execution
            + client_encode_decoding_point
            + client_methods[:-4]  # remove the last 4 white spaces / indentation#
    )
    try:
        with open(config['target_path'], 'wb') as f:
            f.write(all_in_one.encode('utf-8'))

        msg = (f"Done! The client was created at:\n\n{config['target_path']}\n"
               f"Next step is to customize your client and test it with some sample requests.\n")
        print(msg)
    except IOError:
        raise IOError(f'Could not write at the desired location at:\n'
                      f'{config["target_path"]}\n')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--config-path', '-c', action='store', default='config.json')
    args = parser.parse_args()
    main(config_path=args.config_path)



from setuptools import setup, find_packages
import os
import re

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(os.path.join(here, "swagccg", "src", "__init__.py")) as f:
    VERSION = re.search(r"__version__ = '(.+?)'", f.read()).group(1)

name = 'swagccg'
setup(
    name=name,  # Required
    version=VERSION,
    description='Swagger Client Code Generator. Using Python. For Python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/erkandem/swagccg-py2py',
    author='Erkan Demiralay',
    author_email='erkan.dem@pm.me',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

    ],
    keywords='REST api-client client-code code-generator swagger openapi',
    packages=find_packages(exclude=['swagccg.docs', 'swagccg.tests']),
    # install_requires=['peppercorn'],
    # extras_require={  # Optional
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    # },
    # package_data={  # Optional
    #    'sample': ['package_data.dat'],
    # },
    # data_files=[('my_data', ['data/data_file'])],  # Optional
    # entry_points={  # Optional
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    # },
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/erkandem/swagccg-py2py/issues',
        'Source': 'https://github.com/erkandem/swagccg-py2py/',
        'Documentation': 'https://erkandem.github.io/swagccg-py2py/',
    },
)



"""
auto-generated 2020-04-10 18:19:18
... using [swagccg-py2py](https://erkandem.github.io/swagccg-py2py)' version 0.3.3

your module level doc-string goes here
"""

# #######################################################################
# DO NOT MODIFY THIS FILE!
# Your changes will be lost if you rerun ``make_client.py``! 
# Edit the template!
# #######################################################################

from datetime import datetime as dt, timedelta
import json
import urllib
import urllib3
import certifi
import warnings


class MyClientClass(object):
    """your client class level doc-string goes here"""

    def __init__(self, deployment='remote'):
        if deployment == 'remote':
            self.API_PORT = '80'
            self.API_URL_BASE = 'petstore.swagger.io'
            self.API_PROTOCOL = 'https'
        elif deployment == 'local':
            self.API_PORT = '5000'
            self.API_URL_BASE = '127.0.0.1'
            self.API_PROTOCOL = 'http'

        self.BASE_PATH = '/api'
        self.LOGIN_TIMESTAMP = None
        self.API_TOKEN = None
        self.REFRESH_TIMESTAMP = None

        self.AUTH_HEADER_NAME = 'Authorization'
        self.AUTH_PREFIX = 'Bearer '  # mind the whitespace
        self.AUTH_TOKEN_KEY = 'access_token'
        self.AUTH_TOKEN_KEY_REFRESH = 'refreshed_token'
        self.REFRESH_KEY = 'token'

        if self.API_PORT == '80':
            self.API_URL = f'{self.API_PROTOCOL}://{self.API_URL_BASE}'
        else:
            self.API_URL = f'{self.API_PROTOCOL}://{self.API_URL_BASE}:{self.API_PORT}'

        if self.API_PROTOCOL == 'https':
            self.http = urllib3.PoolManager(
                cert_reqs='CERT_REQUIRED',
                ca_certs=certifi.where()
            )
        else:
            self.http = urllib3.PoolManager()

        self.API_LOGIN_URL = f'{self.API_URL}{self.BASE_PATH}/login'
        self.API_REFRESH_URL = f'{self.API_URL}{self.BASE_PATH}/refresh'
        self.API_BASE_URL = f'{self.API_URL}{self.BASE_PATH}'

    def __dir__(self):
        method_names = [
            'get_find_pets_r',
            'post_add_pet_r',
            'get_find_pet_by_id_r',
            'delete_pet_r'
        ]
        return method_names
    
    def login_with_api(self, *, body, headers=None, **kwargs):
        """
        login with the target API and save the JWT token within the class
        
        Args:
            data: login data externally supplied
            body: data to be sent in body (typically credentials)
            headers: option to supply custom headers if needed
        """
        if headers is None:
            headers = {'Content-Type': 'application/json'}
        else:
            if 'content-type' not in [h.lower() for h in headers]:
                headers['Content-Type'] = 'application/json'
        r = self._do_call(
                method='POST',
                url=self.API_LOGIN_URL,
                headers=headers,
                body=body,
                pass_through=True,
                **kwargs
        )
        if r.status == 200:
            res = json.loads(r.data.decode('utf-8'))
            self.API_TOKEN = res[self.AUTH_TOKEN_KEY]
            self.LOGIN_TIMESTAMP = dt.now()
            self.REFRESH_TIMESTAMP = None
        else:
            print(f'login failed \nstatus:{r.status} \n \nurl: {self.API_LOGIN_URL}'
                  '\nIs the username and password correct?')

    # -----------------------------------------------------------------------
    # ---------- Token Management
    # -----------------------------------------------------------------------

    def is_it_time_to_refresh_the_token(self):
        """
        Return True or False depending on the ``LOGIN_TIMESTAMP`` for the
        first refresh or the ``REFRESH_TIMESTAMP`` if the JWT was already
        refreshed once
        
        expiry is server specific
        """
        if self.REFRESH_TIMESTAMP is None:
            if (self.LOGIN_TIMESTAMP + timedelta(hours=10)) < dt.now():
                self.refresh_the_login()
                return True
            else:
                return False
        else:
            if (self.REFRESH_TIMESTAMP + timedelta(hours=10)) < dt.now():
                self.refresh_the_login()
                return True
            else:
                return False

    def refresh_the_login(self):
        """ server specific refresh routine"""
        encoded_data = json.dumps({'token': self.API_TOKEN}).encode('utf-8')
        r = self.http.request(
                'POST',
                self.API_REFRESH_URL,
                headers={'Content-Type': 'application/json'},
                body=encoded_data
        )
        res = json.loads(r.data.decode('utf-8'))
        self.API_TOKEN = res[self.AUTH_TOKEN_KEY_REFRESH]
        self.REFRESH_TIMESTAMP = dt.now()
    
    def _add_auth_header(self, headers=None):
        """ adds the preconfigured authorization header """
        if headers is None:
            headers = dict()
        headers[self.AUTH_HEADER_NAME] = f'{self.AUTH_PREFIX}{self.API_TOKEN}'
        return headers

    def _do_call(self, method=None, url=None, headers=None, fields=None, body=None, **kwargs):
        """
        A way to separate each resource from the actual request dispatching point
        Response is assumed to be json by default. any other mapping can be hooked here.

        Use ``pass_through = True`` to receive the untouched response object

        Args:
            method (str): HTTP-Method
            url (str): endpoint
            headers (dict): each key:value pair represents one header field:value. Don't nest!
            fields (dict):  each key:value pair will be urlencoded and passed as query string. Don't nest!
            body (dict): will be encoded to JSON and bytes afterwards
                         You can get a urlencoding by setting
                         'Content-Type': 'application/x-www-form-urlencoded'

        """

        headers = self._add_auth_header(headers)
        if body is not None and method in ['POST', 'PUT', 'PATCH']:
            if 'Content-Type' not in headers:
                headers['Content-Type'] = 'application/json'
                r = self.http.request(
                        method=method,
                        url=url,
                        body=self._encode(body),
                        headers=headers
                    )
            else:
                if headers['Content-Type'] == 'application/x-www-form-urlencoded':
                    r = self.http.urlopen(
                            method,
                            url,
                            body=self._encode(body, 'url'),
                            headers=headers
                    )
                elif headers['Content-Type'] == 'application/json':
                    r = self.http.request(
                            method=method,
                            url=url,
                            body=self._encode(body),
                            headers=headers
                    )
                else:
                    msg = f''' The Content-Type header was set to {headers['Content-Type']}\n
                    However, anything else than 'application/json' or 'application/x-www-form-urlencoded'\n
                    is not accounted for in the client.\n If you would like to add it look for:\n\n
                    client_point_of_execution_f to build the logic\n
                    client_encoding_decoding_point_f for handling encoding\n\n
                    -1 (negative one) was returned to avoid a RunTimeError'''
                    warnings.warn(msg)
                    return -1
        else:
            r = self.http.request_encode_url(
                    method=method,
                    url=url,
                    headers=headers,
                    fields=fields
            )
        if kwargs.get('pass_through'):
            return r

        if r.status == 200:
            if len(r.data) > 0:
                return self._decode(r.data)
            else:
                return r.status
        elif r.status == 401:
            self.refresh_the_login()
            return 401
        else:
            return -1
    
    def _encode(self, data, format=None):
        """
        Abstracted encoding point. Mount your custom function.
        Focus here is on built in JSON.

        Args:
            data(): python object
            format(str): json or url

        Returns:
            data_encoded: :func:`json.dumps` and encode from utf-8 to binary

        """
        if type(data) is bytes:
            return data
        if format == 'url':
            return (urllib.parse.urlencode(data)).encode('utf-8')
        if format is None:
            return (json.dumps(data)).encode('utf-8')
        elif format == 'json':
            return (json.dumps(data)).encode('utf-8')
        else:
            msg = f"received format = {format}.\nUse 'json' or 'url'.\n 'json' is default."
            raise NotImplementedError(msg)

    def _decode(self, data):
        """
        abstracted decoding point 
        Mount your custom function. Focus here is on JSON.

        Args:
            data: python object (dict, list, ...)

        Returns:
           data_decoded: first decode from binary to utf-8 and parse with 
                         built-in :func:`json.loads`
        """

        return json.loads(data.decode('utf-8')) 
    
    def get_find_pets_r(self, headers=None, body=None, fields_data=None, **kwargs):
        """   """
        r = self._do_call(
                method='GET',
                url=f'{self.API_BASE_URL}/pets',
                headers=headers,
                body=body,
                fields=fields_data,
                **kwargs
        )
        return r
    
    def post_add_pet_r(self, headers=None, body=None, fields_data=None, **kwargs):
        """   """
        r = self._do_call(
                method='POST',
                url=f'{self.API_BASE_URL}/pets',
                headers=headers,
                body=body,
                fields=fields_data,
                **kwargs
        )
        return r
    
    def get_find_pet_by_id_r(self, id, headers=None, body=None, fields_data=None, **kwargs):
        """   """
        r = self._do_call(
                method='GET',
                url=f'{self.API_BASE_URL}/pets/{id}',
                headers=headers,
                body=body,
                fields=fields_data,
                **kwargs
        )
        return r
    
    def delete_pet_r(self, id, headers=None, body=None, fields_data=None, **kwargs):
        """   """
        r = self._do_call(
                method='DELETE',
                url=f'{self.API_BASE_URL}/pets/{id}',
                headers=headers,
                body=body,
                fields=fields_data,
                **kwargs
        )
        return r
